@@ -20,6 +20,7 @@
 
 //! Analysis of regex expressions.
 
+use alloc::boxed::Box;
 use alloc::string::String;
 use alloc::vec::Vec;
 use core::cmp::min;
@@ -28,6 +29,7 @@ use bit_set::BitSet;
 
 use crate::alloc::string::ToString;
 use crate::parse::ExprTree;
+use crate::vm::CaptureGroupRange;
 use crate::{CompileError, Error, Expr, Result};
 
 #[cfg(not(feature = "std"))]
@@ -37,16 +39,28 @@ use std::collections::HashMap as Map;
 
 #[derive(Debug)]
 pub struct Info<'a> {
-    pub(crate) start_group: usize,
-    pub(crate) end_group: usize,
+    pub(crate) capture_groups: CaptureGroupRange,
     pub(crate) min_size: usize,
     pub(crate) const_size: bool,
+    /// Tracks the minimum number of characters that would be consumed in the innermost capture group
+    /// before this expression is matched.
+    pub(crate) min_pos_in_group: usize,
     pub(crate) hard: bool,
     pub(crate) expr: &'a Expr,
     pub(crate) children: Vec<Info<'a>>,
 }
 
 impl<'a> Info<'a> {
+    /// Returns the start (first) group number for this expression.
+    pub(crate) fn start_group(&self) -> usize {
+        self.capture_groups.start()
+    }
+
+    /// Returns the end (last) group number for this expression.
+    pub(crate) fn end_group(&self) -> usize {
+        self.capture_groups.end()
+    }
+
     pub(crate) fn is_literal(&self) -> bool {
         match *self.expr {
             Expr::Literal { casei, .. } => !casei,
@@ -83,7 +97,7 @@ struct Analyzer<'a> {
 }
 
 impl<'a> Analyzer<'a> {
-    fn visit(&mut self, expr: &'a Expr) -> Result<Info<'a>> {
+    fn visit(&mut self, expr: &'a Expr, min_pos_in_group: usize) -> Result<Info<'a>> {
         let start_group = self.group_ix;
         let mut children = Vec::new();
         let mut min_size = 0;
@@ -108,22 +122,24 @@ impl<'a> Analyzer<'a> {
             }
             Expr::Concat(ref v) => {
                 const_size = true;
+                let mut pos_in_group = min_pos_in_group;
                 for child in v {
-                    let child_info = self.visit(child)?;
+                    let child_info = self.visit(child, pos_in_group)?;
                     min_size += child_info.min_size;
                     const_size &= child_info.const_size;
                     hard |= child_info.hard;
+                    pos_in_group += child_info.min_size;
                     children.push(child_info);
                 }
             }
             Expr::Alt(ref v) => {
-                let child_info = self.visit(&v[0])?;
+                let child_info = self.visit(&v[0], min_pos_in_group)?;
                 min_size = child_info.min_size;
                 const_size = child_info.const_size;
                 hard = child_info.hard;
                 children.push(child_info);
                 for child in &v[1..] {
-                    let child_info = self.visit(child)?;
+                    let child_info = self.visit(child, min_pos_in_group)?;
                     const_size &= child_info.const_size && min_size == child_info.min_size;
                     min_size = min(min_size, child_info.min_size);
                     hard |= child_info.hard;
@@ -133,7 +149,7 @@ impl<'a> Analyzer<'a> {
             Expr::Group(ref child) => {
                 let group = self.group_ix;
                 self.group_ix += 1;
-                let child_info = self.visit(child)?;
+                let child_info = self.visit(child, 0)?;
                 min_size = child_info.min_size;
                 const_size = child_info.const_size;
                 // Store the group info for use by backrefs
@@ -151,7 +167,8 @@ impl<'a> Analyzer<'a> {
                 children.push(child_info);
             }
             Expr::LookAround(ref child, _) => {
-                let child_info = self.visit(child)?;
+                // NOTE: min_pos_in_group might seem weird for lookbehinds
+                let child_info = self.visit(child, min_pos_in_group)?;
                 // min_size = 0
                 const_size = true;
                 hard = true;
@@ -160,7 +177,7 @@ impl<'a> Analyzer<'a> {
             Expr::Repeat {
                 ref child, lo, hi, ..
             } => {
-                let child_info = self.visit(child)?;
+                let child_info = self.visit(child, min_pos_in_group)?;
                 min_size = child_info.min_size * lo;
                 const_size = child_info.const_size && lo == hi;
                 hard = child_info.hard;
@@ -173,7 +190,9 @@ impl<'a> Analyzer<'a> {
             }
             Expr::Backref { group, .. } => {
                 if group == 0 {
-                    return Err(Error::CompileError(CompileError::InvalidBackref(group)));
+                    return Err(Error::CompileError(Box::new(CompileError::InvalidBackref(
+                        group,
+                    ))));
                 }
                 // Look up the referenced group's size information
                 if let Some(&SizeInfo {
@@ -187,7 +206,7 @@ impl<'a> Analyzer<'a> {
                 hard = true;
             }
             Expr::AtomicGroup(ref child) => {
-                let child_info = self.visit(child)?;
+                let child_info = self.visit(child, min_pos_in_group)?;
                 min_size = child_info.min_size;
                 const_size = child_info.const_size;
                 hard = true; // TODO: possibly could weaken
@@ -212,9 +231,12 @@ impl<'a> Analyzer<'a> {
             } => {
                 hard = true;
 
-                let child_info_condition = self.visit(condition)?;
-                let child_info_truth = self.visit(true_branch)?;
-                let child_info_false = self.visit(false_branch)?;
+                let child_info_condition = self.visit(condition, min_pos_in_group)?;
+                let child_info_truth = self.visit(
+                    true_branch,
+                    min_pos_in_group + child_info_condition.min_size,
+                )?;
+                let child_info_false = self.visit(false_branch, min_pos_in_group)?;
 
                 min_size = child_info_condition.min_size
                     + min(child_info_truth.min_size, child_info_false.min_size);
@@ -229,18 +251,18 @@ impl<'a> Analyzer<'a> {
                 children.push(child_info_false);
             }
             Expr::SubroutineCall(_) => {
-                return Err(Error::CompileError(CompileError::FeatureNotYetSupported(
-                    "Subroutine Call".to_string(),
+                return Err(Error::CompileError(Box::new(
+                    CompileError::FeatureNotYetSupported("Subroutine Call".to_string()),
                 )));
             }
             Expr::UnresolvedNamedSubroutineCall { ref name, ix } => {
-                return Err(Error::CompileError(
+                return Err(Error::CompileError(Box::new(
                     CompileError::SubroutineCallTargetNotFound(name.to_string(), ix),
-                ));
+                )));
             }
             Expr::BackrefWithRelativeRecursionLevel { .. } => {
-                return Err(Error::CompileError(CompileError::FeatureNotYetSupported(
-                    "Backref at recursion level".to_string(),
+                return Err(Error::CompileError(Box::new(
+                    CompileError::FeatureNotYetSupported("Backref at recursion level".to_string()),
                 )));
             }
         };
@@ -248,11 +270,11 @@ impl<'a> Analyzer<'a> {
         Ok(Info {
             expr,
             children,
-            start_group,
-            end_group: self.group_ix,
+            capture_groups: CaptureGroupRange(start_group, self.group_ix),
             min_size,
             const_size,
             hard,
+            min_pos_in_group,
         })
     }
 }
@@ -273,9 +295,11 @@ pub fn analyze<'a>(tree: &'a ExprTree, explicit_capture_group_0: bool) -> Result
         group_info: Map::new(),
     };
 
-    let analyzed = analyzer.visit(&tree.expr);
+    let analyzed = analyzer.visit(&tree.expr, 0);
     if analyzer.backrefs.contains(0) {
-        return Err(Error::CompileError(CompileError::InvalidBackref(0)));
+        return Err(Error::CompileError(Box::new(CompileError::InvalidBackref(
+            0,
+        ))));
     }
     if let Some(highest_backref) = analyzer.backrefs.into_iter().last() {
         if highest_backref > analyzer.group_ix - start_group
@@ -285,9 +309,9 @@ pub fn analyze<'a>(tree: &'a ExprTree, explicit_capture_group_0: bool) -> Result
             //      `(a(b))\2` has no capture group 2
             || highest_backref == analyzer.group_ix && start_group == 0
         {
-            return Err(Error::CompileError(CompileError::InvalidBackref(
+            return Err(Error::CompileError(Box::new(CompileError::InvalidBackref(
                 highest_backref,
-            )));
+            ))));
         }
     }
     analyzed
@@ -300,11 +324,11 @@ pub fn can_compile_as_anchored(root_expr: &Expr) -> bool {
     use crate::Assertion;
 
     match root_expr {
-        Expr::Concat(ref children) => match children[0] {
-            Expr::Assertion(ref assertion) => *assertion == Assertion::StartText,
+        Expr::Concat(children) => match children[0] {
+            Expr::Assertion(assertion) => assertion == Assertion::StartText,
             _ => false,
         },
-        Expr::Assertion(ref assertion) => *assertion == Assertion::StartText,
+        Expr::Assertion(assertion) => *assertion == Assertion::StartText,
         _ => false,
     }
 }
@@ -335,33 +359,33 @@ mod tests {
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(0)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(0))
         ));
 
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(0)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(0))
         ));
 
         let tree = Expr::parse_tree(r"(.)\0").unwrap();
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(0)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(0))
         ));
 
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(0)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(0))
         ));
 
         let tree = Expr::parse_tree(r"(.)\0\1").unwrap();
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(0)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(0))
         ));
     }
 
@@ -371,14 +395,14 @@ mod tests {
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(1)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(1))
         ));
 
         let tree = Expr::parse_tree(r"aaaa\2").unwrap();
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
     }
 
@@ -388,14 +412,14 @@ mod tests {
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
 
         let tree = Expr::parse_tree(r"a(a)\2\1").unwrap();
         let result = analyze(&tree, false);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
     }
 
@@ -405,28 +429,28 @@ mod tests {
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
 
         let tree = Expr::parse_tree(r"(a(b)\1\2)c").unwrap();
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
 
         let tree = Expr::parse_tree(r"(a\1)b").unwrap();
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(1)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(1))
         ));
 
         let tree = Expr::parse_tree(r"(a(b))\2").unwrap();
         let result = analyze(&tree, true);
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::InvalidBackref(2)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::InvalidBackref(2))
         ));
     }
 
@@ -466,7 +490,7 @@ mod tests {
         assert!(result.is_err());
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::FeatureNotYetSupported(_)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::FeatureNotYetSupported(_))
         ));
 
         let tree = &Expr::parse_tree(r"(a)\k<1-0>").unwrap();
@@ -474,7 +498,7 @@ mod tests {
         assert!(result.is_err());
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(CompileError::FeatureNotYetSupported(_)))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::FeatureNotYetSupported(_))
         ));
     }
 
@@ -485,9 +509,7 @@ mod tests {
         assert!(result.is_err());
         assert!(matches!(
             result.err(),
-            Some(Error::CompileError(
-                CompileError::SubroutineCallTargetNotFound(_, _)
-            ))
+            Some(Error::CompileError(ref box_err)) if matches!(**box_err, CompileError::SubroutineCallTargetNotFound(_, _))
         ));
     }
 
@@ -563,4 +585,66 @@ mod tests {
         let tree = Expr::parse_tree(r"(?m)^(\w+)\1").unwrap();
         assert_eq!(can_compile_as_anchored(&tree.expr), false);
     }
+
+    #[test]
+    fn min_pos_in_group_calculated_correctly_with_no_groups() {
+        let tree = Expr::parse_tree(r"\G").unwrap();
+        let info = analyze(&tree, false).unwrap();
+        assert_eq!(info.min_size, 0);
+        assert_eq!(info.min_pos_in_group, 0);
+        assert!(info.const_size);
+
+        let tree = Expr::parse_tree(r"\G(?=abc)\w+").unwrap();
+        let info = analyze(&tree, false).unwrap();
+        // the lookahead itself has min size 0
+        assert_eq!(info.children[1].min_size, 0);
+        assert!(info.children[1].const_size);
+        // the children of the lookahead have min_size 3 from the literal
+        assert_eq!(info.children[1].children[0].min_size, 3);
+        assert!(info.children[1].children[0].const_size);
+        // after lookahead, the position is reset
+        assert_eq!(info.children[2].min_pos_in_group, 0);
+        assert_eq!(info.children[2].min_size, 1);
+        assert_eq!(info.min_pos_in_group, 0);
+        assert!(!info.const_size);
+
+        let tree = Expr::parse_tree(r"(?:ab*|cd){2}(?=bar)\w").unwrap();
+        let info = analyze(&tree, false).unwrap();
+        // the whole expression has min size 3 (a times 2 plus \w)
+        assert_eq!(info.min_size, 3);
+        // the min pos of the lookahead is 2
+        assert_eq!(info.children[1].min_pos_in_group, 2);
+        // after lookahead, the position is reset
+        assert_eq!(info.children[2].min_pos_in_group, 2);
+        assert_eq!(info.children[2].min_size, 1);
+        assert!(!info.const_size);
+    }
+
+    #[test]
+    fn min_pos_in_group_calculated_correctly_with_capture_groups() {
+        use matches::assert_matches;
+
+        let tree = Expr::parse_tree(r"a(bc)d(e(f)g)").unwrap();
+        let info = analyze(&tree, false).unwrap();
+        assert_eq!(info.min_pos_in_group, 0);
+        // before the capture begins, the min pos in group 0 is 1
+        assert_eq!(info.children[1].min_pos_in_group, 1);
+        // inside capture group 1, the min pos of the Concat inside the group is 0
+        assert_matches!(info.children[1].children[0].expr, Expr::Concat(_));
+        assert_eq!(info.children[1].children[0].min_pos_in_group, 0);
+        assert!(info.children[1].children[0].const_size);
+        // inside capture group 1, the min pos of the c inside the group is 1
+        assert_matches!(info.children[1].children[0].children[1].expr, Expr::Literal { val, casei: false } if val == "c");
+        assert_eq!(info.children[1].children[0].children[1].min_pos_in_group, 1);
+
+        // prove we are looking at the position of the d after capture group 1
+        assert_matches!(info.children[2].expr, Expr::Literal { val, casei: false } if val == "d");
+        assert_eq!(info.children[2].min_pos_in_group, 3);
+        assert_eq!(info.children[2].start_group(), 2);
+        assert_eq!(info.children[2].min_size, 1);
+
+        // prove we are looking at the position of the e in capture group 2
+        assert_matches!(info.children[3].children[0].children[0].expr, Expr::Literal { val, casei: false } if val == "e");
+        assert_eq!(info.children[3].children[0].children[0].min_pos_in_group, 0);
+    }
 }
