@@ -36,8 +36,9 @@
 //! value is longer than 5 characters
 //! ```
 use crate::{
-    paths::Location,
+    paths::{LazyLocation, Location},
     types::{JsonType, JsonTypeSet},
+    validator::LazyEvaluationPath,
 };
 use serde_json::{Map, Number, Value};
 use std::{
@@ -61,7 +62,10 @@ struct ValidationErrorRepr<'a> {
     instance: Cow<'a, Value>,
     kind: ValidationErrorKind,
     instance_path: Location,
+    /// Canonical schema location without $ref traversals (JSON Schema "keywordLocation")
     schema_path: Location,
+    /// Dynamic path including $ref traversals.
+    tracker: LazyEvaluationPath,
 }
 
 /// An iterator over instances of [`ValidationError`] that represent validation error for the
@@ -379,6 +383,7 @@ impl<'a> ValidationError<'a> {
         kind: ValidationErrorKind,
         instance_path: Location,
         schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
     ) -> Self {
         Self {
             repr: Box::new(ValidationErrorRepr {
@@ -386,6 +391,7 @@ impl<'a> ValidationError<'a> {
                 kind,
                 instance_path,
                 schema_path,
+                tracker: tracker.into(),
             }),
         }
     }
@@ -411,23 +417,48 @@ impl<'a> ValidationError<'a> {
         &self.repr.instance_path
     }
 
-    /// Returns the JSON Pointer to the schema keyword that failed validation.
+    /// Returns the canonical schema location without `$ref` traversals.
+    ///
+    /// This corresponds to JSON Schema's "keywordLocation" in output formats.
+    /// See JSON Schema 2020-12 Core, Section 12.4.2.
     #[inline]
     #[must_use]
     pub fn schema_path(&self) -> &Location {
         &self.repr.schema_path
     }
 
-    /// Decomposes the error into owned parts.
+    /// Returns the dynamic evaluation path including `$ref` traversals.
+    ///
+    /// This corresponds to JSON Schema's "evaluationPath" - the actual path taken
+    /// through the schema including by-reference applicators (`$ref`, `$dynamicRef`).
+    /// See JSON Schema 2020-12 Core, Section 12.4.2.
     #[inline]
     #[must_use]
-    pub fn into_parts(self) -> (Cow<'a, Value>, ValidationErrorKind, Location, Location) {
+    pub fn evaluation_path(&self) -> &Location {
+        self.repr.tracker.resolve(&self.repr.schema_path)
+    }
+
+    /// Decomposes the error into owned parts.
+    /// Returns (instance, kind, `instance_path`, `schema_path`, `evaluation_path`).
+    #[inline]
+    #[must_use]
+    pub fn into_parts(
+        self,
+    ) -> (
+        Cow<'a, Value>,
+        ValidationErrorKind,
+        Location,
+        Location,
+        Location,
+    ) {
         let repr = *self.repr;
+        let evaluation_path = repr.tracker.into_owned(repr.schema_path.clone());
         (
             repr.instance,
             repr.kind,
             repr.instance_path,
             repr.schema_path,
+            evaluation_path,
         )
     }
 
@@ -437,8 +468,15 @@ impl<'a> ValidationError<'a> {
         kind: ValidationErrorKind,
         instance_path: Location,
         schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
     ) -> Self {
-        Self::new(Cow::Borrowed(instance), kind, instance_path, schema_path)
+        Self::new(
+            Cow::Borrowed(instance),
+            kind,
+            instance_path,
+            schema_path,
+            tracker,
+        )
     }
 
     /// Returns a wrapper that masks instance values in error messages.
@@ -461,17 +499,19 @@ impl<'a> ValidationError<'a> {
     /// Converts the `ValidationError` into an owned version with `'static` lifetime.
     #[must_use]
     pub fn to_owned(self) -> ValidationError<'static> {
-        let (instance, kind, instance_path, schema_path) = self.into_parts();
+        let (instance, kind, instance_path, schema_path, tracker) = self.into_parts();
         ValidationError::new(
             Cow::Owned(instance.into_owned()),
             kind,
             instance_path,
             schema_path,
+            tracker,
         )
     }
 
     pub(crate) fn additional_items(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: usize,
@@ -480,11 +520,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::AdditionalItems { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn additional_properties(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         unexpected: Vec<String>,
@@ -493,11 +535,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::AdditionalProperties { unexpected },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn any_of(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         context: Vec<Vec<ValidationError<'a>>>,
@@ -511,11 +555,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::AnyOf { context },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn backtrack_limit(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         error: fancy_regex::Error,
@@ -524,11 +570,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::BacktrackLimitExceeded { error },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_array(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         expected_value: &[Value],
@@ -539,11 +587,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::Array(expected_value.to_vec()),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_boolean(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         expected_value: bool,
@@ -554,11 +604,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::Bool(expected_value),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_null(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
     ) -> ValidationError<'a> {
@@ -568,11 +620,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::Null,
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_number(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         expected_value: &Number,
@@ -583,11 +637,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::Number(expected_value.clone()),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_object(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         expected_value: &Map<String, Value>,
@@ -598,11 +654,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::Object(expected_value.clone()),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn constant_string(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         expected_value: &str,
@@ -613,11 +671,13 @@ impl<'a> ValidationError<'a> {
                 expected_value: Value::String(expected_value.to_string()),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn contains(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
     ) -> ValidationError<'a> {
@@ -625,11 +685,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Contains,
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn content_encoding(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         encoding: &str,
@@ -640,11 +702,13 @@ impl<'a> ValidationError<'a> {
                 content_encoding: encoding.to_string(),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn content_media_type(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         media_type: &str,
@@ -655,11 +719,13 @@ impl<'a> ValidationError<'a> {
                 content_media_type: media_type.to_string(),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn enumeration(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         options: &Value,
@@ -670,11 +736,13 @@ impl<'a> ValidationError<'a> {
                 options: options.clone(),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn exclusive_maximum(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: Value,
@@ -683,11 +751,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::ExclusiveMaximum { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn exclusive_minimum(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: Value,
@@ -696,11 +766,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::ExclusiveMinimum { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn false_schema(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
     ) -> ValidationError<'a> {
@@ -708,11 +780,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::FalseSchema,
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn format(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         format: impl Into<String>,
@@ -723,7 +797,8 @@ impl<'a> ValidationError<'a> {
                 format: format.into(),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn from_utf8(error: FromUtf8Error) -> ValidationError<'a> {
@@ -732,10 +807,12 @@ impl<'a> ValidationError<'a> {
             ValidationErrorKind::FromUtf8 { error },
             Location::new(),
             Location::new(),
+            Location::new(),
         )
     }
     pub(crate) fn max_items(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -744,11 +821,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MaxItems { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn maximum(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: Value,
@@ -757,11 +836,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Maximum { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn max_length(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -770,11 +851,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MaxLength { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn max_properties(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -783,11 +866,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MaxProperties { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn min_items(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -796,11 +881,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MinItems { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn minimum(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: Value,
@@ -809,11 +896,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Minimum { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn min_length(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -822,11 +911,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MinLength { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn min_properties(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         limit: u64,
@@ -835,12 +926,14 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MinProperties { limit },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     #[cfg(feature = "arbitrary-precision")]
     pub(crate) fn multiple_of(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         multiple_of: Value,
@@ -849,13 +942,15 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MultipleOf { multiple_of },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
 
     #[cfg(not(feature = "arbitrary-precision"))]
     pub(crate) fn multiple_of(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         multiple_of: f64,
@@ -864,11 +959,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::MultipleOf { multiple_of },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn not(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         schema: Value,
@@ -877,11 +974,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Not { schema },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn one_of_multiple_valid(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         context: Vec<Vec<ValidationError<'a>>>,
@@ -895,11 +994,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::OneOfMultipleValid { context },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn one_of_not_valid(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         context: Vec<Vec<ValidationError<'a>>>,
@@ -913,11 +1014,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::OneOfNotValid { context },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn pattern(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         pattern: String,
@@ -926,11 +1029,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Pattern { pattern },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn property_names(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         error: ValidationError<'a>,
@@ -941,11 +1046,13 @@ impl<'a> ValidationError<'a> {
                 error: Box::new(error.to_owned()),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn required(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         property: Value,
@@ -954,12 +1061,14 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::Required { property },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
 
     pub(crate) fn single_type_error(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         type_name: JsonType,
@@ -970,11 +1079,13 @@ impl<'a> ValidationError<'a> {
                 kind: TypeKind::Single(type_name),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn multiple_type_error(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         types: JsonTypeSet,
@@ -985,11 +1096,13 @@ impl<'a> ValidationError<'a> {
                 kind: TypeKind::Multiple(types),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn unevaluated_items(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         unexpected: Vec<String>,
@@ -998,11 +1111,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::UnevaluatedItems { unexpected },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn unevaluated_properties(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         unexpected: Vec<String>,
@@ -1011,11 +1126,13 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::UnevaluatedProperties { unexpected },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
     pub(crate) fn unique_items(
-        location: Location,
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
     ) -> ValidationError<'a> {
@@ -1023,12 +1140,93 @@ impl<'a> ValidationError<'a> {
             instance,
             ValidationErrorKind::UniqueItems,
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
-    /// Create a new custom validation error.
-    pub fn custom(
-        location: Location,
+    /// Create a custom validation error with just a message.
+    ///
+    /// Use this in [`Keyword::validate`](crate::Keyword::validate) implementations.
+    /// The actual instance, instance path, and schema path are filled in automatically.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// use jsonschema::ValidationError;
+    ///
+    /// fn validate_even(n: u64) -> Result<(), ValidationError<'static>> {
+    ///     if n % 2 != 0 {
+    ///         return Err(ValidationError::custom("number must be even"));
+    ///     }
+    ///     Ok(())
+    /// }
+    /// ```
+    pub fn custom(message: impl Into<String>) -> ValidationError<'static> {
+        ValidationError::new(
+            Cow::Owned(Value::Null),
+            ValidationErrorKind::Custom {
+                message: message.into(),
+            },
+            Location::new(),
+            Location::new(),
+            Location::new(),
+        )
+    }
+
+    /// Create an error for invalid schema values in keyword factories.
+    ///
+    /// Use this in custom keyword factory functions when the schema value
+    /// is invalid for your custom keyword.
+    pub fn schema(message: impl Into<String>) -> ValidationError<'static> {
+        ValidationError::new(
+            Cow::Owned(Value::Null),
+            ValidationErrorKind::Custom {
+                message: message.into(),
+            },
+            Location::new(),
+            Location::new(),
+            Location::new(),
+        )
+    }
+
+    /// Fill in context for a placeholder validation error.
+    ///
+    /// Used by custom keywords, which never involve `$ref` traversal,
+    /// so evaluation path equals schema path.
+    pub(crate) fn with_context<'i>(
+        self,
+        instance: &'i Value,
+        instance_path: &LazyLocation,
+        schema_path: &Location,
+    ) -> ValidationError<'i> {
+        ValidationError::new(
+            Cow::Borrowed(instance),
+            self.repr.kind,
+            instance_path.into(),
+            schema_path.clone(),
+            // Custom keywords are never reached via $ref, so evaluation path = schema path
+            LazyEvaluationPath::SameAsSchemaPath,
+        )
+    }
+
+    /// Fill in context for a placeholder schema error (used in keyword factories).
+    pub(crate) fn with_schema_context(
+        self,
+        schema_value: &Value,
+        schema_path: Location,
+    ) -> ValidationError<'_> {
+        ValidationError::new(
+            Cow::Borrowed(schema_value),
+            self.repr.kind,
+            Location::new(),
+            schema_path.clone(),
+            schema_path,
+        )
+    }
+
+    pub(crate) fn compile_error(
+        schema_path: Location,
+        tracker: impl Into<LazyEvaluationPath>,
         instance_path: Location,
         instance: &'a Value,
         message: impl Into<String>,
@@ -1039,7 +1237,8 @@ impl<'a> ValidationError<'a> {
                 message: message.into(),
             },
             instance_path,
-            location,
+            schema_path,
+            tracker,
         )
     }
 }
@@ -1053,6 +1252,7 @@ impl From<referencing::Error> for ValidationError<'_> {
             ValidationErrorKind::Referencing(err),
             Location::new(),
             Location::new(),
+            Location::new(),
         )
     }
 }
@@ -1527,7 +1727,13 @@ mod tests {
     use test_case::test_case;
 
     fn owned_error(instance: Value, kind: ValidationErrorKind) -> ValidationError<'static> {
-        ValidationError::new(Cow::Owned(instance), kind, Location::new(), Location::new())
+        ValidationError::new(
+            Cow::Owned(instance),
+            kind,
+            Location::new(),
+            Location::new(),
+            Location::new(),
+        )
     }
 
     #[test]
@@ -1690,6 +1896,7 @@ mod tests {
     fn single_type_error() {
         let instance = json!(42);
         let err = ValidationError::single_type_error(
+            Location::new(),
             Location::new(),
             Location::new(),
             &instance,
@@ -1705,6 +1912,7 @@ mod tests {
             .insert(JsonType::String)
             .insert(JsonType::Number);
         let err = ValidationError::multiple_type_error(
+            Location::new(),
             Location::new(),
             Location::new(),
             &instance,
@@ -1872,8 +2080,13 @@ mod tests {
         "value is not of type \"string\""
     )]
     fn test_masked_error_messages(instance: Value, kind: ValidationErrorKind, expected: &str) {
-        let error =
-            ValidationError::new(Cow::Owned(instance), kind, Location::new(), Location::new());
+        let error = ValidationError::new(
+            Cow::Owned(instance),
+            kind,
+            Location::new(),
+            Location::new(),
+            Location::new(),
+        );
         assert_eq!(error.masked().to_string(), expected);
     }
 
@@ -1897,8 +2110,13 @@ mod tests {
         placeholder: &str,
         expected: &str,
     ) {
-        let error =
-            ValidationError::new(Cow::Owned(instance), kind, Location::new(), Location::new());
+        let error = ValidationError::new(
+            Cow::Owned(instance),
+            kind,
+            Location::new(),
+            Location::new(),
+            Location::new(),
+        );
         assert_eq!(error.masked_with(placeholder).to_string(), expected);
     }
 }
