@@ -16,7 +16,7 @@ use crate::{
     compiler, ecma,
     error::ValidationError,
     keywords::CompilationResult,
-    paths::{LazyLocation, Location},
+    paths::{LazyLocation, Location, RefTracker},
     types::JsonType,
     validator::{Validate, ValidationContext},
     Draft,
@@ -733,12 +733,14 @@ macro_rules! format_validators {
                     &self,
                     instance: &'i Value,
                     location: &LazyLocation,
+                    tracker: Option<&RefTracker>,
                     ctx: &mut ValidationContext,
                 ) -> Result<(), ValidationError<'i>> {
                     if let Value::String(_item) = instance {
                         if !self.is_valid(instance, ctx) {
                             return Err(ValidationError::format(
                                 self.location.clone(),
+                                crate::paths::capture_evaluation_path(tracker, &self.location),
                                 location.into(),
                                 instance,
                                 $format,
@@ -813,12 +815,14 @@ impl Validate for EmailValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if let Value::String(_item) = instance {
             if !self.is_valid(instance, ctx) {
                 return Err(ValidationError::format(
                     self.location.clone(),
+                    crate::paths::capture_evaluation_path(tracker, &self.location),
                     location.into(),
                     instance,
                     "email",
@@ -859,12 +863,14 @@ impl Validate for IdnEmailValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if let Value::String(_item) = instance {
             if !self.is_valid(instance, ctx) {
                 return Err(ValidationError::format(
                     self.location.clone(),
+                    crate::paths::capture_evaluation_path(tracker, &self.location),
                     location.into(),
                     instance,
                     "idn-email",
@@ -900,6 +906,7 @@ impl Validate for CustomFormatValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -907,6 +914,7 @@ impl Validate for CustomFormatValidator {
         } else {
             Err(ValidationError::format(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 self.format_name.clone(),
@@ -982,9 +990,11 @@ pub(crate) fn compile<'a>(
                 if ctx.are_unknown_formats_ignored() {
                     None
                 } else {
-                    Some(Err(ValidationError::custom(
-                        Location::new().join("format"),
-                        ctx.location().clone(),
+                    let location = ctx.location().join("format");
+                    Some(Err(ValidationError::compile_error(
+                        location.clone(),
+                        location,
+                        Location::new(),
                         schema,
                         format!("Unknown format: '{name}'. Adjust configuration to ignore unrecognized formats"),
                     )))
@@ -992,9 +1002,11 @@ pub(crate) fn compile<'a>(
             }
         }
     } else {
+        let location = ctx.location().join("format");
         Some(Err(ValidationError::single_type_error(
+            location.clone(),
+            location,
             Location::new(),
-            ctx.location().clone(),
             schema,
             JsonType::String,
         )))
