@@ -3,7 +3,7 @@ use crate::{
     error::ValidationError,
     ext::numeric,
     keywords::CompilationResult,
-    paths::{LazyLocation, Location},
+    paths::{LazyLocation, Location, RefTracker},
     types::JsonType,
     validator::{Validate, ValidationContext},
 };
@@ -37,6 +37,7 @@ macro_rules! define_numeric_keywords {
                     &self,
                     instance: &'i Value,
                     location: &LazyLocation,
+                    tracker: Option<&RefTracker>,
                     ctx: &mut ValidationContext,
                 ) -> Result<(), ValidationError<'i>> {
                     if self.is_valid(instance, ctx) {
@@ -44,6 +45,7 @@ macro_rules! define_numeric_keywords {
                     } else {
                         Err(ValidationError::$error_fn_name(
                             self.location.clone(),
+                            crate::paths::capture_evaluation_path(tracker, &self.location),
                             location.into(),
                             instance,
                             self.limit_val.clone(),
@@ -73,7 +75,8 @@ define_numeric_keywords!(
 #[cfg(feature = "arbitrary-precision")]
 pub(crate) mod bigint_validators {
     use super::{
-        numeric, LazyLocation, Location, Validate, ValidationContext, ValidationError, Value,
+        numeric, LazyLocation, Location, RefTracker, Validate, ValidationContext, ValidationError,
+        Value,
     };
     use crate::ext::numeric::bignum::{
         f64_ge_bigfrac, f64_ge_bigint, f64_gt_bigfrac, f64_gt_bigint, f64_le_bigfrac,
@@ -104,6 +107,7 @@ pub(crate) mod bigint_validators {
                     &self,
                     instance: &'i Value,
                     location: &LazyLocation,
+                    tracker: Option<&RefTracker>,
                     ctx: &mut ValidationContext,
                 ) -> Result<(), ValidationError<'i>> {
                     if self.is_valid(instance, ctx) {
@@ -111,6 +115,7 @@ pub(crate) mod bigint_validators {
                     } else {
                         Err(ValidationError::$error_fn(
                             self.location.clone(),
+                            crate::paths::capture_evaluation_path(tracker, &self.location),
                             location.into(),
                             instance,
                             self.limit_val.clone(),
@@ -118,7 +123,7 @@ pub(crate) mod bigint_validators {
                     }
                 }
 
-                fn is_valid(&self, instance: &Value, _ctx: &mut ValidationContext) -> bool {
+                fn is_valid(&self, instance: &Value, _ctx: &mut  ValidationContext) -> bool {
                     use fraction::BigFraction;
                     if let Value::Number(item) = instance {
                         // Try to parse instance as BigInt first
@@ -215,6 +220,7 @@ pub(crate) mod bigint_validators {
                     &self,
                     instance: &'i Value,
                     location: &LazyLocation,
+                    tracker: Option<&RefTracker>,
                     ctx: &mut ValidationContext,
                 ) -> Result<(), ValidationError<'i>> {
                     if self.is_valid(instance, ctx) {
@@ -222,6 +228,7 @@ pub(crate) mod bigint_validators {
                     } else {
                         Err(ValidationError::$error_fn(
                             self.location.clone(),
+                            crate::paths::capture_evaluation_path(tracker, &self.location),
                             location.into(),
                             instance,
                             self.limit_val.clone(),
@@ -229,7 +236,7 @@ pub(crate) mod bigint_validators {
                     }
                 }
 
-                fn is_valid(&self, instance: &Value, _ctx: &mut ValidationContext) -> bool {
+                fn is_valid(&self, instance: &Value, _ctx: &mut  ValidationContext) -> bool {
                     if let Value::Number(item) = instance {
                         // Try to parse instance as BigFraction for exact precision
                         if let Some(instance_bigfrac) = try_parse_bigfraction(item) {
@@ -304,10 +311,16 @@ where
     Ok(Box::new(V::from((limit, schema.clone(), location))))
 }
 
-fn number_type_error<'a>(ctx: &compiler::Context, schema: &'a Value) -> CompilationResult<'a> {
+fn number_type_error<'a>(
+    ctx: &compiler::Context,
+    keyword: &str,
+    schema: &'a Value,
+) -> CompilationResult<'a> {
+    let location = ctx.location().join(keyword);
     Err(ValidationError::single_type_error(
+        location.clone(),
+        location,
         Location::new(),
-        ctx.location().clone(),
         schema,
         JsonType::Number,
     ))
@@ -429,7 +442,7 @@ pub(crate) fn compile_minimum<'a>(
 ) -> Option<CompilationResult<'a>> {
     match schema {
         Value::Number(limit) => create_numeric_validator!(Minimum, ctx, "minimum", limit, schema),
-        _ => Some(number_type_error(ctx, schema)),
+        _ => Some(number_type_error(ctx, "minimum", schema)),
     }
 }
 
@@ -441,7 +454,7 @@ pub(crate) fn compile_maximum<'a>(
 ) -> Option<CompilationResult<'a>> {
     match schema {
         Value::Number(limit) => create_numeric_validator!(Maximum, ctx, "maximum", limit, schema),
-        _ => Some(number_type_error(ctx, schema)),
+        _ => Some(number_type_error(ctx, "maximum", schema)),
     }
 }
 
@@ -455,7 +468,7 @@ pub(crate) fn compile_exclusive_minimum<'a>(
         Value::Number(limit) => {
             create_numeric_validator!(ExclusiveMinimum, ctx, "exclusiveMinimum", limit, schema)
         }
-        _ => Some(number_type_error(ctx, schema)),
+        _ => Some(number_type_error(ctx, "exclusiveMinimum", schema)),
     }
 }
 
@@ -469,7 +482,7 @@ pub(crate) fn compile_exclusive_maximum<'a>(
         Value::Number(limit) => {
             create_numeric_validator!(ExclusiveMaximum, ctx, "exclusiveMaximum", limit, schema)
         }
-        _ => Some(number_type_error(ctx, schema)),
+        _ => Some(number_type_error(ctx, "exclusiveMaximum", schema)),
     }
 }
 
