@@ -1,15 +1,16 @@
 use crate::{
     compiler,
     error::ValidationError,
+    evaluation::ErrorDescription,
     keywords::CompilationResult,
     paths::Location,
     types::{JsonType, JsonTypeSet},
-    validator::{Validate, ValidationContext},
+    validator::{EvaluationResult, Validate, ValidationContext},
 };
 use serde_json::{json, Map, Number, Value};
 use std::str::FromStr;
 
-use crate::paths::LazyLocation;
+use crate::paths::{LazyLocation, RefTracker};
 
 pub(crate) struct MultipleTypesValidator {
     types: JsonTypeSet,
@@ -27,8 +28,9 @@ impl MultipleTypesValidator {
                         types = types.insert(ty);
                     } else {
                         return Err(ValidationError::enumeration(
-                            Location::new(),
+                            location.clone(),
                             location,
+                            Location::new(),
                             item,
                             &json!([
                                 "array", "boolean", "integer", "null", "number", "object", "string"
@@ -38,6 +40,7 @@ impl MultipleTypesValidator {
                 }
                 _ => {
                     return Err(ValidationError::single_type_error(
+                        location.clone(),
                         location,
                         Location::new(),
                         item,
@@ -58,6 +61,7 @@ impl Validate for MultipleTypesValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -65,12 +69,27 @@ impl Validate for MultipleTypesValidator {
         } else {
             Err(ValidationError::multiple_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 self.types,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            let message = format!("{instance} is not of types {:?}", self.types);
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new("type", message)])
+        }
+    }
 }
 
 pub(crate) struct NullTypeValidator {
@@ -92,6 +111,7 @@ impl Validate for NullTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -99,12 +119,29 @@ impl Validate for NullTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Null,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "null""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct BooleanTypeValidator {
@@ -126,6 +163,7 @@ impl Validate for BooleanTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -133,12 +171,29 @@ impl Validate for BooleanTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Boolean,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "boolean""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct StringTypeValidator {
@@ -161,6 +216,7 @@ impl Validate for StringTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -168,12 +224,29 @@ impl Validate for StringTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::String,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "string""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct ArrayTypeValidator {
@@ -196,6 +269,7 @@ impl Validate for ArrayTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -203,12 +277,29 @@ impl Validate for ArrayTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Array,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "array""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct ObjectTypeValidator {
@@ -230,6 +321,7 @@ impl Validate for ObjectTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -237,12 +329,29 @@ impl Validate for ObjectTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Object,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "object""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct NumberTypeValidator {
@@ -264,6 +373,7 @@ impl Validate for NumberTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -271,12 +381,29 @@ impl Validate for NumberTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Number,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "number""#),
+            )])
+        }
+    }
 }
 
 pub(crate) struct IntegerTypeValidator {
@@ -302,6 +429,7 @@ impl Validate for IntegerTypeValidator {
         &self,
         instance: &'i Value,
         location: &LazyLocation,
+        tracker: Option<&RefTracker>,
         ctx: &mut ValidationContext,
     ) -> Result<(), ValidationError<'i>> {
         if self.is_valid(instance, ctx) {
@@ -309,12 +437,29 @@ impl Validate for IntegerTypeValidator {
         } else {
             Err(ValidationError::single_type_error(
                 self.location.clone(),
+                crate::paths::capture_evaluation_path(tracker, &self.location),
                 location.into(),
                 instance,
                 JsonType::Integer,
             ))
         }
     }
+    fn evaluate(
+        &self,
+        instance: &Value,
+        _location: &LazyLocation,
+        _tracker: Option<&RefTracker>,
+        ctx: &mut ValidationContext,
+    ) -> EvaluationResult {
+        if self.is_valid(instance, ctx) {
+            EvaluationResult::valid_empty()
+        } else {
+            EvaluationResult::invalid_empty(vec![ErrorDescription::new(
+                "type",
+                format!(r#"{instance} is not of type "integer""#),
+            )])
+        }
+    }
 }
 
 fn is_integer(num: &Number) -> bool {
@@ -369,8 +514,9 @@ pub(crate) fn compile<'a>(
                     Some(compile_single_type(ty.as_str(), location, item))
                 } else {
                     Some(Err(ValidationError::single_type_error(
-                        Location::new(),
+                        location.clone(),
                         location,
+                        Location::new(),
                         item,
                         JsonType::String,
                     )))
@@ -379,14 +525,18 @@ pub(crate) fn compile<'a>(
                 Some(MultipleTypesValidator::compile(items, location))
             }
         }
-        _ => Some(Err(ValidationError::multiple_type_error(
-            Location::new(),
-            ctx.location().clone(),
-            schema,
-            JsonTypeSet::empty()
-                .insert(JsonType::String)
-                .insert(JsonType::Array),
-        ))),
+        _ => {
+            let location = ctx.location().join("type");
+            Some(Err(ValidationError::multiple_type_error(
+                location.clone(),
+                location,
+                Location::new(),
+                schema,
+                JsonTypeSet::empty()
+                    .insert(JsonType::String)
+                    .insert(JsonType::Array),
+            )))
+        }
     }
 }
 
@@ -403,9 +553,10 @@ fn compile_single_type<'a>(
         Ok(JsonType::Number) => NumberTypeValidator::compile(location),
         Ok(JsonType::Object) => ObjectTypeValidator::compile(location),
         Ok(JsonType::String) => StringTypeValidator::compile(location),
-        Err(()) => Err(ValidationError::custom(
-            Location::new(),
+        Err(()) => Err(ValidationError::compile_error(
+            location.clone(),
             location,
+            Location::new(),
             instance,
             "Unexpected type",
         )),
