@@ -1,4 +1,10 @@
-use crate::{compiler, node::SchemaNode, paths::Location, validator::Validate as _};
+use crate::{
+    compiler,
+    node::SchemaNode,
+    paths::{LazyEvaluationPath, Location},
+    validator::Validate as _,
+    ValidationContext,
+};
 use ahash::AHashMap;
 use serde_json::{Map, Value};
 
@@ -91,12 +97,12 @@ pub(crate) fn compile_big_map<'a>(
 pub(crate) fn are_properties_valid<M, F>(
     prop_map: &M,
     props: &Map<String, Value>,
-    ctx: &mut crate::validator::ValidationContext,
+    ctx: &mut ValidationContext,
     check: F,
 ) -> bool
 where
     M: PropertiesValidatorsMap,
-    F: Fn(&Value, &mut crate::validator::ValidationContext) -> bool,
+    F: Fn(&Value, &mut ValidationContext) -> bool,
 {
     for (property, instance) in props {
         if let Some(validator) = prop_map.get_validator(property) {
@@ -121,7 +127,13 @@ pub(crate) fn compile_fancy_regex_patterns<'a>(
     for (pattern, subschema) in obj {
         let pctx = kctx.new_at_location(pattern.as_str());
         let compiled_pattern = ctx.get_or_compile_regex(pattern).map_err(|()| {
-            ValidationError::format(Location::new(), kctx.location().clone(), subschema, "regex")
+            ValidationError::format(
+                kctx.location().clone(),
+                LazyEvaluationPath::SameAsSchemaPath,
+                Location::new(),
+                subschema,
+                "regex",
+            )
         })?;
         let node = compiler::compile(&pctx, pctx.as_resource_ref(subschema))?;
         compiled_patterns.push(((*compiled_pattern).clone(), node));
@@ -139,7 +151,13 @@ pub(crate) fn compile_regex_patterns<'a>(
     for (pattern, subschema) in obj {
         let pctx = kctx.new_at_location(pattern.as_str());
         let compiled_pattern = ctx.get_or_compile_standard_regex(pattern).map_err(|()| {
-            ValidationError::format(Location::new(), kctx.location().clone(), subschema, "regex")
+            ValidationError::format(
+                kctx.location().clone(),
+                LazyEvaluationPath::SameAsSchemaPath,
+                Location::new(),
+                subschema,
+                "regex",
+            )
         })?;
         let node = compiler::compile(&pctx, pctx.as_resource_ref(subschema))?;
         compiled_patterns.push(((*compiled_pattern).clone(), node));
@@ -148,20 +166,22 @@ pub(crate) fn compile_regex_patterns<'a>(
 }
 
 macro_rules! compile_dynamic_prop_map_validator {
-    ($validator:tt, $properties:ident, $( $arg:expr ),* $(,)*) => {{
+    ($validator:tt, $properties:ident, $ctx:expr, $( $arg:expr ),* $(,)*) => {{
         if let Value::Object(map) = $properties {
             if map.len() < 40 {
                 Some($validator::<SmallValidatorsMap>::compile(
-                    map, $($arg, )*
+                    map, $ctx, $($arg, )*
                 ))
             } else {
                 Some($validator::<BigValidatorsMap>::compile(
-                    map, $($arg, )*
+                    map, $ctx, $($arg, )*
                 ))
             }
         } else {
-            Some(Err(ValidationError::custom(
-                Location::new(),
+            let location = $ctx.location().clone();
+            Some(Err(ValidationError::compile_error(
+                location.clone(),
+                location,
                 Location::new(),
                 $properties,
                 "Unexpected type",
