@@ -4,7 +4,7 @@ use nix::sys::socket::{
     getsockopt, setsockopt, socket, sockopt, AddressFamily, SockFlag,
     SockProtocol, SockType,
 };
-use rand::{thread_rng, Rng};
+use rand::{rng, Rng};
 use std::os::unix::io::{AsRawFd, FromRawFd, OwnedFd};
 
 // NB: FreeBSD supports LOCAL_PEERCRED for SOCK_SEQPACKET, but OSX does not.
@@ -66,6 +66,36 @@ pub fn test_local_peer_pid() {
     assert_eq!(pid, std::process::id() as _);
 }
 
+#[cfg(apple_targets)]
+#[test]
+pub fn test_local_peer_token() {
+    use nix::sys::socket::{audit_token_t, socketpair};
+
+    #[link(name = "bsm", kind = "dylib")]
+    extern "C" {
+        /// Extract the process ID from an `audit_token_t`, used to identify
+        /// Mach tasks and senders of Mach messages as subjects of the audit
+        /// system.
+        ///
+        /// - `atoken`: The Mach audit token.
+        /// - Returns: The process ID extracted from the Mach audit token.
+        fn audit_token_to_pid(atoken: audit_token_t) -> libc::pid_t;
+    }
+
+    let (fd1, _fd2) = socketpair(
+        AddressFamily::Unix,
+        SockType::Stream,
+        None,
+        SockFlag::empty(),
+    )
+    .unwrap();
+    let audit_token = getsockopt(&fd1, sockopt::LocalPeerToken).unwrap();
+    assert_eq!(
+        unsafe { audit_token_to_pid(audit_token) },
+        std::process::id() as _
+    );
+}
+
 #[cfg(target_os = "linux")]
 #[test]
 fn is_so_mark_functional() {
@@ -94,7 +124,7 @@ fn test_so_buf() {
         SockProtocol::Udp,
     )
     .unwrap();
-    let bufsize: usize = thread_rng().gen_range(4096..131_072);
+    let bufsize: usize = rng().random_range(4096..131_072);
     setsockopt(&fd, sockopt::SndBuf, &bufsize).unwrap();
     let actual = getsockopt(&fd, sockopt::SndBuf).unwrap();
     assert!(actual >= bufsize);
@@ -129,11 +159,11 @@ fn test_so_listen_q_limit() {
 }
 
 #[test]
+#[cfg_attr(target_os = "cygwin", ignore)]
 fn test_so_tcp_maxseg() {
     use nix::sys::socket::{
         accept, bind, connect, getsockname, listen, Backlog, SockaddrIn,
     };
-    use nix::unistd::write;
     use std::net::SocketAddrV4;
     use std::str::FromStr;
 
@@ -153,14 +183,12 @@ fn test_so_tcp_maxseg() {
     let initial = getsockopt(&rsock, sockopt::TcpMaxSeg).unwrap();
     // Initial MSS is expected to be 536 (https://tools.ietf.org/html/rfc879#section-1) but some
     // platforms keep it even lower. This might fail if you've tuned your initial MSS to be larger
-    // than 700
+    // than `segsize`
+    let segsize: u32 = 873;
+    assert!(initial < segsize);
     cfg_if! {
         if #[cfg(linux_android)] {
-            let segsize: u32 = 873;
-            assert!(initial < segsize);
             setsockopt(&rsock, sockopt::TcpMaxSeg, &segsize).unwrap();
-        } else {
-            assert!(initial < 700);
         }
     }
 
@@ -172,20 +200,38 @@ fn test_so_tcp_maxseg() {
         SockProtocol::Tcp,
     )
     .unwrap();
+
     connect(ssock.as_raw_fd(), &sock_addr).unwrap();
+
     let rsess = accept(rsock.as_raw_fd()).unwrap();
     let rsess = unsafe { OwnedFd::from_raw_fd(rsess) };
-    write(&rsess, b"hello").unwrap();
-    let actual = getsockopt(&ssock, sockopt::TcpMaxSeg).unwrap();
-    // Actual max segment size takes header lengths into account, max IPv4 options (60 bytes) + max
-    // TCP options (40 bytes) are subtracted from the requested maximum as a lower boundary.
+
     cfg_if! {
-        if #[cfg(linux_android)] {
-            assert!((segsize - 100) <= actual);
-            assert!(actual <= segsize);
+        if #[cfg(apple_targets)] {
+            // on apple targets (and unlike linux), we can only set the MSS on a *connected*
+            // socket. Also, the same MSS can't be read using getsockopt from the other end.
+
+            assert_ne!(segsize, getsockopt(&rsess, sockopt::TcpMaxSeg).unwrap());
+            setsockopt(&rsess, sockopt::TcpMaxSeg, &segsize).unwrap();
+            assert_eq!(segsize, getsockopt(&rsess, sockopt::TcpMaxSeg).unwrap());
+
+            assert_ne!(segsize, getsockopt(&ssock, sockopt::TcpMaxSeg).unwrap());
+            setsockopt(&ssock, sockopt::TcpMaxSeg, &segsize).unwrap();
+            assert_eq!(segsize, getsockopt(&ssock, sockopt::TcpMaxSeg).unwrap());
         } else {
-            assert!(initial < actual);
-            assert!(536 < actual);
+            use nix::unistd::write;
+
+            write(&rsess, b"hello").unwrap();
+            let actual = getsockopt(&ssock, sockopt::TcpMaxSeg).unwrap();
+            // Actual max segment size takes header lengths into account, max IPv4 options (60 bytes) + max
+            // TCP options (40 bytes) are subtracted from the requested maximum as a lower boundary.
+            if cfg!(linux_android) {
+                assert!((segsize - 100) <= actual);
+                assert!(actual <= segsize);
+            } else {
+                assert!(initial < actual);
+                assert!(536 < actual);
+            }
         }
     }
 }
@@ -211,6 +257,8 @@ fn test_so_type_unknown() {
     use nix::errno::Errno;
 
     require_capability!("test_so_type", CAP_NET_RAW);
+    // SOCK_PACKET is deprecated, but since it is used for testing here, we allow it
+    #[allow(deprecated)]
     let raw_fd = unsafe { libc::socket(libc::AF_PACKET, libc::SOCK_PACKET, 0) };
     assert!(raw_fd >= 0, "Error opening socket: {}", nix::Error::last());
     let sockfd = unsafe { OwnedFd::from_raw_fd(raw_fd) };
@@ -219,16 +267,16 @@ fn test_so_type_unknown() {
 }
 
 // The CI doesn't supported getsockopt and setsockopt on emulated processors.
-// It's believed that a QEMU issue, the tests run ok on a fully emulated system.
-// Current CI just run the binary with QEMU but the Kernel remains the same as the host.
+// It's believed to be a QEMU issue; the tests run ok on a fully emulated
+// system.  Current CI just runs the binary with QEMU but the kernel remains the
+// same as the host.
 // So the syscall doesn't work properly unless the kernel is also emulated.
 #[test]
-#[cfg(all(
-    any(target_arch = "x86", target_arch = "x86_64"),
-    any(target_os = "freebsd", target_os = "linux")
-))]
+#[cfg(any(target_os = "freebsd", target_os = "linux"))]
+#[cfg_attr(qemu, ignore)]
 fn test_tcp_congestion() {
     use std::ffi::OsString;
+    use std::os::unix::ffi::OsStrExt;
 
     let fd = socket(
         AddressFamily::Inet,
@@ -239,6 +287,10 @@ fn test_tcp_congestion() {
     .unwrap();
 
     let val = getsockopt(&fd, sockopt::TcpCongestion).unwrap();
+    let bytes = val.as_os_str().as_bytes();
+    for b in bytes.iter() {
+        assert_ne!(*b, 0, "OsString should contain no embedded NULs: {val:?}");
+    }
     setsockopt(&fd, sockopt::TcpCongestion, &val).unwrap();
 
     setsockopt(
@@ -251,6 +303,35 @@ fn test_tcp_congestion() {
     assert_eq!(getsockopt(&fd, sockopt::TcpCongestion).unwrap(), val);
 }
 
+#[test]
+#[cfg(target_os = "freebsd")]
+fn test_tcp_function_blk_alias() {
+    use std::ffi::CStr;
+
+    let fd = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+
+    let tfs = getsockopt(&fd, sockopt::TcpFunctionBlk).unwrap();
+    let name = unsafe { CStr::from_ptr(tfs.function_set_name.as_ptr()) };
+    assert!(!name.to_bytes().is_empty());
+
+    let aliastfs = getsockopt(&fd, sockopt::TcpFunctionAlias).unwrap();
+    let aliasname =
+        unsafe { CStr::from_ptr(aliastfs.function_set_name.as_ptr()) };
+    // freebsd default tcp stack has no alias.
+    assert!(aliasname.to_bytes().is_empty());
+
+    // We can't know at compile time what options are available.  So just test the setter by a
+    // no-op set.
+    // TODO: test if we can load for example BBR tcp stack kernel module.
+    setsockopt(&fd, sockopt::TcpFunctionBlk, &tfs).unwrap();
+}
+
 #[test]
 #[cfg(linux_android)]
 fn test_bindtodevice() {
@@ -420,7 +501,7 @@ fn test_dontfrag_opts() {
 }
 
 #[test]
-#[cfg(any(linux_android, apple_targets))]
+#[cfg(any(all(linux_android, not(target_env = "uclibc")), apple_targets))]
 // Disable the test under emulation because it fails in Cirrus-CI.  Lack
 // of QEMU support is suspected.
 #[cfg_attr(qemu, ignore)]
@@ -469,7 +550,7 @@ fn test_so_priority() {
 }
 
 #[test]
-#[cfg(target_os = "linux")]
+#[cfg(any(linux_android, target_os = "freebsd"))]
 fn test_ip_tos() {
     let fd = socket(
         AddressFamily::Inet,
@@ -479,12 +560,12 @@ fn test_ip_tos() {
     )
     .unwrap();
     let tos = 0x80; // CS4
-    setsockopt(&fd, sockopt::IpTos, &tos).unwrap();
-    assert_eq!(getsockopt(&fd, sockopt::IpTos).unwrap(), tos);
+    setsockopt(&fd, sockopt::Ipv4Tos, &tos).unwrap();
+    assert_eq!(getsockopt(&fd, sockopt::Ipv4Tos).unwrap(), tos);
 }
 
 #[test]
-#[cfg(target_os = "linux")]
+#[cfg(any(linux_android, target_os = "freebsd"))]
 // Disable the test under emulation because it fails in Cirrus-CI.  Lack
 // of QEMU support is suspected.
 #[cfg_attr(qemu, ignore)]
@@ -615,7 +696,7 @@ fn test_ts_clock_monotonic() {
 }
 
 #[test]
-#[cfg(linux_android)]
+#[cfg(all(linux_android, not(target_env = "uclibc")))]
 // Disable the test under emulation because it fails with ENOPROTOOPT in CI
 // on cross target. Lack of QEMU support is suspected.
 #[cfg_attr(qemu, ignore)]
@@ -683,6 +764,46 @@ fn can_get_peercred_on_unix_socket() {
     assert_ne!(a_cred.pid(), 0);
 }
 
+#[cfg(target_os = "linux")]
+fn pid_from_pidfd(pidfd: OwnedFd) -> u32 {
+    use std::fs::read_to_string;
+
+    let fd = pidfd.as_raw_fd();
+    let fdinfo = read_to_string(format!("/proc/self/fdinfo/{fd}")).unwrap();
+    let pidline = fdinfo.split('\n').find(|s| s.starts_with("Pid:")).unwrap();
+    pidline.split('\t').next_back().unwrap().parse().unwrap()
+}
+
+#[cfg(target_os = "linux")]
+#[test]
+fn can_get_peerpidfd_on_unix_socket() {
+    use nix::sys::socket::{socketpair, sockopt, SockFlag, SockType};
+
+    let (a, b) = socketpair(
+        AddressFamily::Unix,
+        SockType::Stream,
+        None,
+        SockFlag::empty(),
+    )
+    .unwrap();
+
+    match (
+        getsockopt(&a, sockopt::PeerPidfd),
+        getsockopt(&b, sockopt::PeerPidfd),
+    ) {
+        (Ok(a_pidfd), Ok(b_pidfd)) => {
+            let a_pid = pid_from_pidfd(a_pidfd);
+            let b_pid = pid_from_pidfd(b_pidfd);
+            assert_eq!(a_pid, b_pid);
+            assert_ne!(a_pid, 0);
+        }
+        (Err(nix::Error::ENOPROTOOPT), Err(nix::Error::ENOPROTOOPT)) => {
+            // Pidfd can still be unsupported on some CI runners
+        }
+        (Err(err), _) | (_, Err(err)) => panic!("{err:?}"),
+    };
+}
+
 #[test]
 fn is_socket_type_unix() {
     use nix::sys::socket::{socketpair, sockopt, SockFlag, SockType};
@@ -877,3 +998,275 @@ fn test_reuseport_lb() {
     setsockopt(&fd, sockopt::ReusePortLb, &true).unwrap();
     assert!(getsockopt(&fd, sockopt::ReusePortLb).unwrap());
 }
+
+#[test]
+#[cfg(any(linux_android, target_os = "freebsd"))]
+fn test_ipv4_recv_ttl_opts() {
+    let fd = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        SockProtocol::Tcp,
+    )
+    .unwrap();
+    setsockopt(&fd, sockopt::Ipv4RecvTtl, &true)
+        .expect("setting IP_RECVTTL on an inet stream socket should succeed");
+    setsockopt(&fd, sockopt::Ipv4RecvTtl, &false)
+        .expect("unsetting IP_RECVTTL on an inet stream socket should succeed");
+    let fdd = socket(
+        AddressFamily::Inet,
+        SockType::Datagram,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    setsockopt(&fdd, sockopt::Ipv4RecvTtl, &true)
+        .expect("setting IP_RECVTTL on an inet datagram socket should succeed");
+    setsockopt(&fdd, sockopt::Ipv4RecvTtl, &false).expect(
+        "unsetting IP_RECVTTL on an inet datagram socket should succeed",
+    );
+}
+
+#[test]
+#[cfg(any(linux_android, target_os = "freebsd"))]
+fn test_ipv6_recv_hop_limit_opts() {
+    let fd = socket(
+        AddressFamily::Inet6,
+        SockType::Stream,
+        SockFlag::empty(),
+        SockProtocol::Tcp,
+    )
+    .unwrap();
+    setsockopt(&fd, sockopt::Ipv6RecvHopLimit, &true).expect(
+        "setting IPV6_RECVHOPLIMIT on an inet6 stream socket should succeed",
+    );
+    setsockopt(&fd, sockopt::Ipv6RecvHopLimit, &false).expect(
+        "unsetting IPV6_RECVHOPLIMIT on an inet6 stream socket should succeed",
+    );
+    let fdd = socket(
+        AddressFamily::Inet6,
+        SockType::Datagram,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    setsockopt(&fdd, sockopt::Ipv6RecvHopLimit, &true).expect(
+        "setting IPV6_RECVHOPLIMIT on an inet6 datagram socket should succeed",
+    );
+    setsockopt(&fdd, sockopt::Ipv6RecvHopLimit, &false).expect(
+        "unsetting IPV6_RECVHOPLIMIT on an inet6 datagram socket should succeed",
+    );
+}
+
+#[test]
+#[cfg(any(linux_android, target_os = "freebsd"))]
+fn test_ipv4_recv_tos_opts() {
+    let fd = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        SockProtocol::Tcp,
+    )
+    .unwrap();
+    setsockopt(&fd, sockopt::IpRecvTos, &true)
+        .expect("setting IP_RECVTOS on an inet stream socket should succeed");
+    setsockopt(&fd, sockopt::IpRecvTos, &false)
+        .expect("unsetting IP_RECVTOS on an inet stream socket should succeed");
+    let fdd = socket(
+        AddressFamily::Inet,
+        SockType::Datagram,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    setsockopt(&fdd, sockopt::IpRecvTos, &true)
+        .expect("setting IP_RECVTOS on an inet datagram socket should succeed");
+    setsockopt(&fdd, sockopt::IpRecvTos, &false).expect(
+        "unsetting IP_RECVTOS on an inet datagram socket should succeed",
+    );
+}
+
+#[test]
+#[cfg(any(linux_android, target_os = "freebsd"))]
+fn test_ipv6_recv_traffic_class_opts() {
+    let fd = socket(
+        AddressFamily::Inet6,
+        SockType::Stream,
+        SockFlag::empty(),
+        SockProtocol::Tcp,
+    )
+    .unwrap();
+    setsockopt(&fd, sockopt::Ipv6RecvTClass, &true).expect(
+        "setting IPV6_RECVTCLASS on an inet6 stream socket should succeed",
+    );
+    setsockopt(&fd, sockopt::Ipv6RecvTClass, &false).expect(
+        "unsetting IPV6_RECVTCLASS on an inet6 stream socket should succeed",
+    );
+    let fdd = socket(
+        AddressFamily::Inet6,
+        SockType::Datagram,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    setsockopt(&fdd, sockopt::Ipv6RecvTClass, &true).expect(
+        "setting IPV6_RECVTCLASS on an inet6 datagram socket should succeed",
+    );
+    setsockopt(&fdd, sockopt::Ipv6RecvTClass, &false).expect(
+        "unsetting IPV6_RECVTCLASS on an inet6 datagram socket should succeed",
+    );
+}
+
+#[cfg(apple_targets)]
+#[test]
+fn test_linger_sec() {
+    let fd = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+
+    let set_linger = libc::linger {
+        l_onoff: 1,
+        l_linger: 1,
+    };
+    setsockopt(&fd, sockopt::LingerSec, &set_linger).unwrap();
+
+    let get_linger = getsockopt(&fd, sockopt::Linger).unwrap();
+    assert_eq!(get_linger.l_linger, set_linger.l_linger * 100);
+}
+
+/// Users should be able to define their own sockopts.
+mod sockopt_impl {
+    use nix::sys::socket::{
+        getsockopt, setsockopt, socket, AddressFamily, SockFlag, SockProtocol,
+        SockType,
+    };
+
+    sockopt_impl!(KeepAlive, Both, libc::SOL_SOCKET, libc::SO_KEEPALIVE, bool);
+
+    #[test]
+    fn test_so_tcp_keepalive() {
+        let fd = socket(
+            AddressFamily::Inet,
+            SockType::Stream,
+            SockFlag::empty(),
+            SockProtocol::Tcp,
+        )
+        .unwrap();
+        setsockopt(&fd, KeepAlive, &true).unwrap();
+        assert!(getsockopt(&fd, KeepAlive).unwrap());
+    }
+
+    sockopt_impl!(
+        Linger,
+        Both,
+        libc::SOL_SOCKET,
+        libc::SO_LINGER,
+        libc::linger
+    );
+    #[test]
+    fn test_linger() {
+        let fd = socket(
+            AddressFamily::Inet,
+            SockType::Stream,
+            SockFlag::empty(),
+            None,
+        )
+        .unwrap();
+
+        let set_linger = libc::linger {
+            l_onoff: 1,
+            l_linger: 42,
+        };
+        setsockopt(&fd, Linger, &set_linger).unwrap();
+
+        let get_linger = getsockopt(&fd, Linger).unwrap();
+        assert_eq!(get_linger.l_linger, set_linger.l_linger);
+    }
+}
+
+#[cfg(solarish)]
+#[test]
+fn test_exclbind() {
+    use nix::errno::Errno;
+    use nix::sys::socket::{
+        bind, socket, AddressFamily, SockFlag, SockType, SockaddrIn,
+    };
+    use std::net::SocketAddrV4;
+    use std::str::FromStr;
+    let fd1 = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    let addr = SocketAddrV4::from_str("127.0.0.1:8081").unwrap();
+    let excl = true;
+
+    setsockopt(&fd1, sockopt::ExclBind, &excl).unwrap();
+    bind(fd1.as_raw_fd(), &SockaddrIn::from(addr)).unwrap();
+    assert_eq!(getsockopt(&fd1, sockopt::ExclBind), Ok(true));
+    let fd2 = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    assert_eq!(
+        bind(fd2.as_raw_fd(), &SockaddrIn::from(addr)),
+        Err(Errno::EADDRINUSE)
+    );
+}
+
+#[cfg(target_os = "illumos")]
+#[test]
+fn test_solfilter() {
+    use nix::errno::Errno;
+    let s = socket(
+        AddressFamily::Inet,
+        SockType::Stream,
+        SockFlag::empty(),
+        SockProtocol::Tcp,
+    )
+    .unwrap();
+    let data = std::ffi::OsStr::new("httpf");
+    let attach = sockopt::FilterAttach;
+    let detach = sockopt::FilterDetach;
+
+    // These 2 options won't work unless the needed kernel module is installed:
+    // https://github.com/nix-rust/nix/pull/2611#issuecomment-2750237782
+    //
+    // So we only test the binding here
+    assert_eq!(Err(Errno::ENOENT), setsockopt(&s, attach, data));
+    assert_eq!(Err(Errno::ENOENT), setsockopt(&s, detach, data));
+}
+
+#[cfg(target_os = "linux")]
+#[test]
+pub fn test_so_attach_reuseport_cbpf() {
+    let fd = socket(
+        AddressFamily::Inet6,
+        SockType::Datagram,
+        SockFlag::empty(),
+        None,
+    )
+    .unwrap();
+    setsockopt(&fd, sockopt::ReusePort, &true).unwrap();
+    setsockopt(&fd, sockopt::ReuseAddr, &true).unwrap();
+    let mut flt: [libc::sock_filter; 2] = unsafe { std::mem::zeroed() };
+    flt[0].code = (libc::BPF_LD | libc::BPF_W | libc::BPF_ABS) as u16;
+    flt[0].k = (libc::SKF_AD_OFF + libc::SKF_AD_CPU) as u32;
+    flt[1].code = (libc::BPF_RET | 0x10) as u16;
+    let fp = libc::sock_fprog {
+        len: flt.len() as u16,
+        filter: flt.as_mut_ptr(),
+    };
+    setsockopt(&fd, sockopt::AttachReusePortCbpf, &fp).unwrap_or_else(|e| {
+        assert_eq!(e, nix::errno::Errno::ENOPROTOOPT);
+    });
+}
