@@ -9,9 +9,9 @@ use multimap::MultiMap;
 use prost_types::field_descriptor_proto::{Label, Type};
 use prost_types::source_code_info::Location;
 use prost_types::{
-    DescriptorProto, EnumDescriptorProto, EnumValueDescriptorProto, FieldDescriptorProto,
-    FieldOptions, FileDescriptorProto, OneofDescriptorProto, ServiceDescriptorProto,
-    SourceCodeInfo,
+    DescriptorProto, EnumDescriptorProto, EnumValueDescriptorProto, EnumValueOptions,
+    FieldDescriptorProto, FieldOptions, FileDescriptorProto, OneofDescriptorProto,
+    ServiceDescriptorProto, SourceCodeInfo,
 };
 
 use crate::ast::{Comments, Method, Service};
@@ -65,20 +65,42 @@ struct OneofField {
     descriptor: OneofDescriptorProto,
     fields: Vec<Field>,
     path_index: i32,
+    // This type has the same name as another nested type at the same level
+    has_type_name_conflict: bool,
 }
 
 impl OneofField {
-    fn new(descriptor: OneofDescriptorProto, fields: Vec<Field>, path_index: i32) -> Self {
+    fn new(
+        parent: &DescriptorProto,
+        descriptor: OneofDescriptorProto,
+        fields: Vec<Field>,
+        path_index: i32,
+    ) -> Self {
+        let nested_type_names = parent.nested_type.iter().map(DescriptorProto::name);
+        let nested_enum_names = parent.enum_type.iter().map(EnumDescriptorProto::name);
+        let has_type_name_conflict = nested_type_names
+            .chain(nested_enum_names)
+            .any(|type_name| to_upper_camel(type_name) == to_upper_camel(descriptor.name()));
+
         Self {
             descriptor,
             fields,
             path_index,
+            has_type_name_conflict,
         }
     }
 
     fn rust_name(&self) -> String {
         to_snake(self.descriptor.name())
     }
+
+    fn type_name(&self) -> String {
+        let mut name = to_upper_camel(self.descriptor.name());
+        if self.has_type_name_conflict {
+            name.push_str("OneOf");
+        }
+        name
+    }
 }
 
 impl<'b> CodeGenerator<'_, 'b> {
@@ -167,7 +189,7 @@ impl<'b> CodeGenerator<'_, 'b> {
         type MapTypes = HashMap<String, (FieldDescriptorProto, FieldDescriptorProto)>;
         let (nested_types, map_types): (NestedTypes, MapTypes) = message
             .nested_type
-            .into_iter()
+            .iter()
             .enumerate()
             .partition_map(|(idx, nested_type)| {
                 if nested_type
@@ -184,7 +206,7 @@ impl<'b> CodeGenerator<'_, 'b> {
                     let name = format!("{}.{}", &fq_message_name, nested_type.name());
                     Either::Right((name, (key, value)))
                 } else {
-                    Either::Left((nested_type, idx))
+                    Either::Left((nested_type.clone(), idx))
                 }
             });
 
@@ -193,28 +215,28 @@ impl<'b> CodeGenerator<'_, 'b> {
         type OneofFieldsByIndex = MultiMap<i32, Field>;
         let (fields, mut oneof_map): (Vec<Field>, OneofFieldsByIndex) = message
             .field
-            .into_iter()
+            .iter()
             .enumerate()
             .partition_map(|(idx, proto)| {
                 let idx = idx as i32;
                 if proto.proto3_optional.unwrap_or(false) {
-                    Either::Left(Field::new(proto, idx))
+                    Either::Left(Field::new(proto.clone(), idx))
                 } else if let Some(oneof_index) = proto.oneof_index {
-                    Either::Right((oneof_index, Field::new(proto, idx)))
+                    Either::Right((oneof_index, Field::new(proto.clone(), idx)))
                 } else {
-                    Either::Left(Field::new(proto, idx))
+                    Either::Left(Field::new(proto.clone(), idx))
                 }
             });
         // Optional fields create a synthetic oneof that we want to skip
         let oneof_fields: Vec<OneofField> = message
             .oneof_decl
-            .into_iter()
+            .iter()
             .enumerate()
-            .filter_map(move |(idx, proto)| {
+            .filter_map(|(idx, proto)| {
                 let idx = idx as i32;
                 oneof_map
                     .remove(&idx)
-                    .map(|fields| OneofField::new(proto, fields, idx))
+                    .map(|fields| OneofField::new(&message, proto.clone(), fields, idx))
             })
             .collect();
 
@@ -236,6 +258,7 @@ impl<'b> CodeGenerator<'_, 'b> {
             },
             self.context.prost_path()
         ));
+        self.append_prost_path_attribute();
         self.append_skip_debug(&fq_message_name);
         self.push_indent();
         self.buf.push_str("pub struct ");
@@ -310,10 +333,8 @@ impl<'b> CodeGenerator<'_, 'b> {
         ));
         self.depth += 1;
 
-        self.buf.push_str(&format!(
-            "const NAME: &'static str = \"{}\";\n",
-            message_name,
-        ));
+        self.buf
+            .push_str(&format!("const NAME: &'static str = \"{message_name}\";\n"));
         self.buf.push_str(&format!(
             "const PACKAGE: &'static str = \"{}\";\n",
             self.package,
@@ -360,6 +381,13 @@ impl<'b> CodeGenerator<'_, 'b> {
         }
     }
 
+    fn append_prost_path_attribute(&mut self) {
+        if let Some(prost_path_attribute) = self.context.prost_path_attribute() {
+            self.buf.push_str(prost_path_attribute);
+            self.buf.push('\n');
+        }
+    }
+
     fn append_skip_debug(&mut self, fq_message_name: &str) {
         if self.context.should_skip_debug(fq_message_name) {
             push_indent(self.buf, self.depth);
@@ -420,7 +448,7 @@ impl<'b> CodeGenerator<'_, 'b> {
                 .context
                 .bytes_type(fq_message_name, field.descriptor.name());
             self.buf
-                .push_str(&format!("={:?}", bytes_type.annotation()));
+                .push_str(&format!(" = {:?}", bytes_type.annotation()));
         }
 
         match field.descriptor.label() {
@@ -439,7 +467,7 @@ impl<'b> CodeGenerator<'_, 'b> {
                         .as_ref()
                         .map_or(self.syntax == Syntax::Proto3, |options| options.packed())
                 {
-                    self.buf.push_str(", packed=\"false\"");
+                    self.buf.push_str(", packed = \"false\"");
                 }
             }
         }
@@ -447,11 +475,11 @@ impl<'b> CodeGenerator<'_, 'b> {
         if boxed {
             self.buf.push_str(", boxed");
         }
-        self.buf.push_str(", tag=\"");
+        self.buf.push_str(", tag = \"");
         self.buf.push_str(&field.descriptor.number().to_string());
 
         if let Some(ref default) = field.descriptor.default_value {
-            self.buf.push_str("\", default=\"");
+            self.buf.push_str("\", default = \"");
             if type_ == Type::Bytes {
                 self.buf.push_str("b\\\"");
                 for b in unescape_c_escape_string(default) {
@@ -492,13 +520,13 @@ impl<'b> CodeGenerator<'_, 'b> {
 
         if repeated {
             self.buf
-                .push_str(&format!("{}::alloc::vec::Vec<", prost_path));
+                .push_str(&format!("{prost_path}::alloc::vec::Vec<"));
         } else if optional {
             self.buf.push_str("::core::option::Option<");
         }
         if boxed {
             self.buf
-                .push_str(&format!("{}::alloc::boxed::Box<", prost_path));
+                .push_str(&format!("{prost_path}::alloc::boxed::Box<"));
         }
         self.buf.push_str(&ty);
         if boxed {
@@ -537,7 +565,7 @@ impl<'b> CodeGenerator<'_, 'b> {
         let value_tag = self.map_value_type_tag(value);
 
         self.buf.push_str(&format!(
-            "#[prost({}=\"{}, {}\", tag=\"{}\")]\n",
+            "#[prost({} = \"{}, {}\", tag = \"{}\")]\n",
             map_type.annotation(),
             key_tag,
             value_tag,
@@ -545,13 +573,25 @@ impl<'b> CodeGenerator<'_, 'b> {
         ));
         self.append_field_attributes(fq_message_name, field.descriptor.name());
         self.push_indent();
-        self.buf.push_str(&format!(
-            "pub {}: {}<{}, {}>,\n",
-            field.rust_name(),
-            map_type.rust_type(),
-            key_ty,
-            value_ty
-        ));
+        match map_type {
+            crate::MapType::HashMap => {
+                self.buf.push_str(&format!(
+                    "pub {}: ::std::collections::HashMap<{}, {}>,\n",
+                    field.rust_name(),
+                    key_ty,
+                    value_ty
+                ));
+            }
+            crate::MapType::BTreeMap => {
+                self.buf.push_str(&format!(
+                    "pub {}: {}::alloc::collections::BTreeMap<{}, {}>,\n",
+                    field.rust_name(),
+                    self.context.prost_path(),
+                    key_ty,
+                    value_ty
+                ));
+            }
+        }
     }
 
     fn append_oneof_field(
@@ -560,15 +600,11 @@ impl<'b> CodeGenerator<'_, 'b> {
         fq_message_name: &str,
         oneof: &OneofField,
     ) {
-        let type_name = format!(
-            "{}::{}",
-            to_snake(message_name),
-            to_upper_camel(oneof.descriptor.name())
-        );
+        let type_name = format!("{}::{}", to_snake(message_name), oneof.type_name());
         self.append_doc(fq_message_name, None);
         self.push_indent();
         self.buf.push_str(&format!(
-            "#[prost(oneof=\"{}\", tags=\"{}\")]\n",
+            "#[prost(oneof = \"{}\", tags = \"{}\")]\n",
             type_name,
             oneof
                 .fields
@@ -615,10 +651,11 @@ impl<'b> CodeGenerator<'_, 'b> {
             },
             self.context.prost_path()
         ));
+        self.append_prost_path_attribute();
         self.append_skip_debug(fq_message_name);
         self.push_indent();
         self.buf.push_str("pub enum ");
-        self.buf.push_str(&to_upper_camel(oneof.descriptor.name()));
+        self.buf.push_str(&oneof.type_name());
         self.buf.push_str(" {\n");
 
         self.path.push(2);
@@ -628,10 +665,15 @@ impl<'b> CodeGenerator<'_, 'b> {
             self.append_doc(fq_message_name, Some(field.descriptor.name()));
             self.path.pop();
 
+            if self.deprecated(&field.descriptor) {
+                self.push_indent();
+                self.buf.push_str("#[deprecated]\n");
+            }
+
             self.push_indent();
             let ty_tag = self.field_type_tag(&field.descriptor);
             self.buf.push_str(&format!(
-                "#[prost({}, tag=\"{}\")]\n",
+                "#[prost({}, tag = \"{}\")]\n",
                 ty_tag,
                 field.descriptor.number()
             ));
@@ -655,8 +697,9 @@ impl<'b> CodeGenerator<'_, 'b> {
 
             if boxed {
                 self.buf.push_str(&format!(
-                    "{}(::prost::alloc::boxed::Box<{}>),\n",
+                    "{}({}::alloc::boxed::Box<{}>),\n",
                     to_upper_camel(field.descriptor.name()),
+                    self.context.prost_path(),
                     ty
                 ));
             } else {
@@ -722,6 +765,7 @@ impl<'b> CodeGenerator<'_, 'b> {
             dbg,
             self.context.prost_path(),
         ));
+        self.append_prost_path_attribute();
         self.push_indent();
         self.buf.push_str("#[repr(i32)]\n");
         self.push_indent();
@@ -739,6 +783,10 @@ impl<'b> CodeGenerator<'_, 'b> {
 
             self.append_doc(&fq_proto_enum_name, Some(variant.proto_name));
             self.append_field_attributes(&fq_proto_enum_name, variant.proto_name);
+            if variant.deprecated {
+                self.push_indent();
+                self.buf.push_str("#[deprecated]\n");
+            }
             self.push_indent();
             self.buf.push_str(&variant.generated_variant_name);
             self.buf.push_str(" = ");
@@ -785,6 +833,10 @@ impl<'b> CodeGenerator<'_, 'b> {
         self.depth += 1;
 
         for variant in variant_mappings.iter() {
+            if variant.deprecated {
+                self.push_indent();
+                self.buf.push_str("#[allow(deprecated)]\n");
+            }
             self.push_indent();
             self.buf.push_str("Self::");
             self.buf.push_str(&variant.generated_variant_name);
@@ -818,7 +870,11 @@ impl<'b> CodeGenerator<'_, 'b> {
             self.push_indent();
             self.buf.push('\"');
             self.buf.push_str(variant.proto_name);
-            self.buf.push_str("\" => Some(Self::");
+            self.buf.push_str("\" => Some(");
+            if variant.deprecated {
+                self.buf.push_str("#[allow(deprecated)] ");
+            }
+            self.buf.push_str("Self::");
             self.buf.push_str(&variant.generated_variant_name);
             self.buf.push_str("),\n");
         }
@@ -841,7 +897,7 @@ impl<'b> CodeGenerator<'_, 'b> {
 
     fn push_service(&mut self, service: ServiceDescriptorProto) {
         let name = service.name().to_owned();
-        debug!("  service: {:?}", name);
+        debug!("  service: {name:?}");
 
         let comments = self
             .location()
@@ -940,11 +996,12 @@ impl<'b> CodeGenerator<'_, 'b> {
             Type::Int64 | Type::Sfixed64 | Type::Sint64 => String::from("i64"),
             Type::Bool => String::from("bool"),
             Type::String => format!("{}::alloc::string::String", self.context.prost_path()),
-            Type::Bytes => self
-                .context
-                .bytes_type(fq_message_name, field.name())
-                .rust_type()
-                .to_owned(),
+            Type::Bytes => match self.context.bytes_type(fq_message_name, field.name()) {
+                crate::BytesType::Vec => {
+                    format!("{}::alloc::vec::Vec<u8>", self.context.prost_path())
+                }
+                crate::BytesType::Bytes => format!("{}::bytes::Bytes", self.context.prost_path()),
+            },
             Type::Group | Type::Message => self.resolve_ident(field.type_name()),
         }
     }
@@ -1007,7 +1064,7 @@ impl<'b> CodeGenerator<'_, 'b> {
             Type::Group => Cow::Borrowed("group"),
             Type::Message => Cow::Borrowed("message"),
             Type::Enum => Cow::Owned(format!(
-                "enumeration={:?}",
+                "enumeration = {:?}",
                 self.resolve_ident(field.type_name())
             )),
         }
@@ -1082,6 +1139,7 @@ struct EnumVariantMapping<'a> {
     proto_name: &'a str,
     proto_number: i32,
     generated_variant_name: String,
+    deprecated: bool,
 }
 
 fn build_enum_value_mappings<'a>(
@@ -1117,7 +1175,15 @@ fn build_enum_value_mappings<'a>(
             proto_name: value.name(),
             proto_number: value.number(),
             generated_variant_name,
+            deprecated: enum_field_deprecated(value),
         })
     }
     mappings
 }
+
+fn enum_field_deprecated(value: &EnumValueDescriptorProto) -> bool {
+    value
+        .options
+        .as_ref()
+        .is_some_and(EnumValueOptions::deprecated)
+}
